# coding: utf-8
"""
Collection of builder functions
"""
from typing import Callable, Optional, Generator

import torch
from torch import nn
from torch.optim.lr_scheduler import _LRScheduler, ReduceLROnPlateau, \
    StepLR, ExponentialLR
from torch.optim import Optimizer

from joeynmt.helpers import ConfigurationError


def build_gradient_clipper(config: dict) -> Optional[Callable]:
    """
    Define the function for gradient clipping as specified in configuration.
    If not specified, returns None.

    Current options:
        - "clip_grad_val": clip the gradients if they exceed this value,
            see `torch.nn.utils.clip_grad_value_`
        - "clip_grad_norm": clip the gradients if their norm exceeds this value,
            see `torch.nn.utils.clip_grad_norm_`

    :param config: dictionary with training configurations
    :return: clipping function (in-place) or None if no gradient clipping
    """
    clip_grad_fun = None
    if "clip_grad_val" in config.keys():
        clip_value = config["clip_grad_val"]
        clip_grad_fun = lambda params: \
            nn.utils.clip_grad_value_(parameters=params,
                                      clip_value=clip_value)
    elif "clip_grad_norm" in config.keys():
        max_norm = config["clip_grad_norm"]
        clip_grad_fun = lambda params: \
            nn.utils.clip_grad_norm_(parameters=params, max_norm=max_norm)

    if "clip_grad_val" in config.keys() and "clip_grad_norm" in config.keys():
        raise ConfigurationError(
            "You can only specify either clip_grad_val or clip_grad_norm.")

    return clip_grad_fun


def build_optimizer(config: dict, parameters: Generator) -> Optimizer:
    """
    Create an optimizer for the given parameters as specified in config.

    Except for the weight decay and initial learning rate,
    default optimizer settings are used.

    Currently supported configuration settings for "optimizer":
        - "sgd" (default): see `torch.optim.SGD`
        - "adam": see `torch.optim.adam`
        - "adagrad": see `torch.optim.adagrad`
        - "adadelta": see `torch.optim.adadelta`
        - "rmsprop": see `torch.optim.RMSprop`

    The initial learning rate is set according to "learning_rate" in the config.
    The weight decay is set according to "weight_decay" in the config.
    If they are not specified, the initial learning rate is set to 3.0e-4, the
    weight decay to 0.

    Note that the scheduler state is saved in the checkpoint, so if you load
    a model for further training you have to use the same type of scheduler.

    :param config: configuration dictionary
    :param parameters:
    :return: optimizer
    """
    optimizer_name = config.get("optimizer", "sgd").lower()
    learning_rate = config.get("learning_rate", 3.0e-4)
    weight_decay = config.get("weight_decay", 0)

    if optimizer_name == "adam":
        adam_betas = config.get("adam_betas", (0.9, 0.999))
        optimizer = torch.optim.Adam(parameters, weight_decay=weight_decay,
                                     lr=learning_rate, betas=adam_betas)
    elif optimizer_name == "adagrad":
        optimizer = torch.optim.Adagrad(parameters, weight_decay=weight_decay,
                                        lr=learning_rate)
    elif optimizer_name == "adadelta":
        optimizer = torch.optim.Adadelta(parameters, weight_decay=weight_decay,
                                         lr=learning_rate)
    elif optimizer_name == "rmsprop":
        optimizer = torch.optim.RMSprop(parameters, weight_decay=weight_decay,
                                        lr=learning_rate)
    elif optimizer_name == "sgd":
        # default
        optimizer = torch.optim.SGD(parameters, weight_decay=weight_decay,
                                    lr=learning_rate)
    else:
        raise ConfigurationError("Invalid optimizer. Valid options: 'adam', "
                                 "'adagrad', 'adadelta', 'rmsprop', 'sgd'.")
    return optimizer


def build_scheduler(config: dict, optimizer: Optimizer, scheduler_mode: str,
                    hidden_size: int = 0) \
        -> (Optional[_LRScheduler], Optional[str]):
    """
    Create a learning rate scheduler if specified in config and
    determine when a scheduler step should be executed.

    Current options:
        - "plateau": see `torch.optim.lr_scheduler.ReduceLROnPlateau`
        - "decaying": see `torch.optim.lr_scheduler.StepLR`
        - "exponential": see `torch.optim.lr_scheduler.ExponentialLR`
        - "noam": see `joeynmt.builders.NoamScheduler`
        - "warmupexponentialdecay": see
          `joeynmt.builders.WarmupExponentialDecayScheduler`

    If no scheduler is specified, returns (None, None) which will result in
    a constant learning rate.

    :param config: training configuration
    :param optimizer: optimizer for the scheduler, determines the set of
        parameters which the scheduler sets the learning rate for
    :param scheduler_mode: "min" or "max", depending on whether the validation
        score should be minimized or maximized.
        Only relevant for "plateau".
    :param hidden_size: encoder hidden size (required for NoamScheduler)
    :return:
        - scheduler: scheduler object,
        - scheduler_step_at: either "validation" or "epoch"
    """
    scheduler, scheduler_step_at = None, None
    if "scheduling" in config.keys() and \
            config["scheduling"]:
        if config["scheduling"].lower() == "plateau":
            # learning rate scheduler
            scheduler = ReduceLROnPlateau(
                optimizer=optimizer,
                mode=scheduler_mode,
                verbose=False,
                threshold_mode='abs',
                factor=config.get("decrease_factor", 0.1),
                patience=config.get("patience", 10))
            # scheduler step is executed after every validation
            scheduler_step_at = "validation"
        elif config["scheduling"].lower() == "decaying":
            scheduler = StepLR(
                optimizer=optimizer,
                step_size=config.get("decaying_step_size", 1))
            # scheduler step is executed after every epoch
            scheduler_step_at = "epoch"
        elif config["scheduling"].lower() == "exponential":
            scheduler = ExponentialLR(
                optimizer=optimizer,
                gamma=config.get("decrease_factor", 0.99))
            # scheduler step is executed after every epoch
            scheduler_step_at = "epoch"
        elif config["scheduling"].lower() == "noam":
            factor = config.get("learning_rate_factor", 1)
            warmup = config.get("learning_rate_warmup", 4000)
            scheduler = NoamScheduler(hidden_size=hidden_size, factor=factor,
                                      warmup=warmup, optimizer=optimizer)

            scheduler_step_at = "step"
        elif config["scheduling"].lower() == "warmupexponentialdecay":
            min_rate = config.get("learning_rate_min", 1.0e-5)
            decay_rate = config.get("learning_rate_decay", 0.1)
            warmup = config.get("learning_rate_warmup", 4000)
            peak_rate = config.get("learning_rate_peak", 1.0e-3)
            decay_length = config.get("learning_rate_decay_length", 10000)
            scheduler = WarmupExponentialDecayScheduler(
                min_rate=min_rate, decay_rate=decay_rate,
                warmup=warmup, optimizer=optimizer, peak_rate=peak_rate,
                decay_length=decay_length)
            scheduler_step_at = "step"
    return scheduler, scheduler_step_at


class NoamScheduler:
    """
    The Noam learning rate scheduler used in "Attention is all you need"
    See Eq. 3 in https://arxiv.org/pdf/1706.03762.pdf
    """

    def __init__(self, hidden_size: int, optimizer: torch.optim.Optimizer,
                 factor: float = 1, warmup: int = 4000):
        """
        Warm-up, followed by learning rate decay.

        :param hidden_size:
        :param optimizer:
        :param factor: decay factor
        :param warmup: number of warmup steps
        """
        self.optimizer = optimizer
        self._step = 0
        self.warmup = warmup
        self.factor = factor
        self.hidden_size = hidden_size
        self._rate = 0

    def step(self):
        """Update parameters and rate"""
        self._step += 1
        rate = self._compute_rate()
        for p in self.optimizer.param_groups:
            p['lr'] = rate
        self._rate = rate

    def _compute_rate(self):
        """Implement `lrate` above"""
        step = self._step
        return self.factor * \
            (self.hidden_size ** (-0.5) *
                min(step ** (-0.5), step * self.warmup ** (-1.5)))

    #pylint: disable=no-self-use
    def state_dict(self):
        return None


class WarmupExponentialDecayScheduler:
    """
    A learning rate scheduler similar to Noam, but modified:
    Keep the warm up period but make it so that the decay rate can be tuneable.
    The decay is exponential up to a given minimum rate.
    """

    def __init__(self, optimizer: torch.optim.Optimizer,
                 peak_rate: float = 1.0e-3,
                 decay_length: int = 10000, warmup: int = 4000,
                 decay_rate: float = 0.5, min_rate: float = 1.0e-5):
        """
        Warm-up, followed by exponential learning rate decay.

        :param peak_rate: maximum learning rate at peak after warmup
        :param optimizer:
        :param decay_length: decay length after warmup
        :param decay_rate: decay rate after warmup
        :param warmup: number of warmup steps
        :param min_rate: minimum learning rate
        """
        self.optimizer = optimizer
        self._step = 0
        self.warmup = warmup
        self.decay_length = decay_length
        self.peak_rate = peak_rate
        self._rate = 0
        self.decay_rate = decay_rate
        self.min_rate = min_rate

    def step(self):
        """Update parameters and rate"""
        self._step += 1
        rate = self._compute_rate()
        for p in self.optimizer.param_groups:
            p['lr'] = rate
        self._rate = rate

    def _compute_rate(self):
        """Implement `lrate` above"""
        step = self._step
        warmup = self.warmup

        if step < warmup:
            rate = step * self.peak_rate / warmup
        else:
            exponent = (step - warmup) / self.decay_length
            rate = self.peak_rate * (self.decay_rate ** exponent)
        return max(rate, self.min_rate)

    #pylint: disable=no-self-use
    def state_dict(self):
        return None



import unittest

import torch


class TensorTestCase(unittest.TestCase):
    def assertTensorNotEqual(self, expected, actual):
        equal = torch.equal(expected, actual)
        if equal:
            self.fail("Tensors did match but weren't supposed to: expected {},"
                      " actual {}.".format(expected, actual))

    def assertTensorEqual(self, expected, actual):
        equal = torch.equal(expected, actual)
        if not equal:
            self.fail("Tensors didn't match but were supposed to {} vs"
                      " {}".format(expected, actual))

    def assertTensorAlmostEqual(self, expected, actual):
        diff = torch.all(
            torch.lt(torch.abs(torch.add(expected, -actual)), 1e-4))
        if not diff:
            self.fail("Tensors didn't match but were supposed to {} vs"
                      " {}".format(expected, actual))



# coding: utf-8

"""
Implementation of a mini-batch.
"""


class Batch:
    """Object for holding a batch of data with mask during training.
    Input is a batch from a torch text iterator.
    """

    def __init__(self, torch_batch, pad_index, use_cuda=False):
        """
        Create a new joey batch from a torch batch.
        This batch extends torch text's batch attributes with src and trg
        length, masks, number of non-padded tokens in trg.
        Furthermore, it can be sorted by src length.

        :param torch_batch:
        :param pad_index:
        :param use_cuda:
        """
        self.src, self.src_lengths = torch_batch.src
        self.src_mask = (self.src != pad_index).unsqueeze(1)
        self.nseqs = self.src.size(0)
        self.trg_input = None
        self.trg = None
        self.trg_mask = None
        self.trg_lengths = None
        self.ntokens = None
        self.use_cuda = use_cuda

        if hasattr(torch_batch, "trg"):
            trg, trg_lengths = torch_batch.trg
            # trg_input is used for teacher forcing, last one is cut off
            self.trg_input = trg[:, :-1]
            self.trg_lengths = trg_lengths
            # trg is used for loss computation, shifted by one since BOS
            self.trg = trg[:, 1:]
            # we exclude the padded areas from the loss computation
            self.trg_mask = (self.trg_input != pad_index).unsqueeze(1)
            self.ntokens = (self.trg != pad_index).data.sum().item()

        if use_cuda:
            self._make_cuda()

    def _make_cuda(self):
        """
        Move the batch to GPU

        :return:
        """
        self.src = self.src.cuda()
        self.src_mask = self.src_mask.cuda()

        if self.trg_input is not None:
            self.trg_input = self.trg_input.cuda()
            self.trg = self.trg.cuda()
            self.trg_mask = self.trg_mask.cuda()

    def sort_by_src_lengths(self):
        """
        Sort by src length (descending) and return index to revert sort

        :return:
        """
        _, perm_index = self.src_lengths.sort(0, descending=True)
        rev_index = [0]*perm_index.size(0)
        for new_pos, old_pos in enumerate(perm_index.cpu().numpy()):
            rev_index[old_pos] = new_pos

        sorted_src_lengths = self.src_lengths[perm_index]
        sorted_src = self.src[perm_index]
        sorted_src_mask = self.src_mask[perm_index]
        if self.trg_input is not None:
            sorted_trg_input = self.trg_input[perm_index]
            sorted_trg_lengths = self.trg_lengths[perm_index]
            sorted_trg_mask = self.trg_mask[perm_index]
            sorted_trg = self.trg[perm_index]

        self.src = sorted_src
        self.src_lengths = sorted_src_lengths
        self.src_mask = sorted_src_mask

        if self.trg_input is not None:
            self.trg_input = sorted_trg_input
            self.trg_mask = sorted_trg_mask
            self.trg_lengths = sorted_trg_lengths
            self.trg = sorted_trg

        if self.use_cuda:
            self._make_cuda()

        return rev_index



#!/usr/bin/env python
from setuptools import setup, find_packages
import sys
import os

with open("requirements.txt", encoding="utf-8") as req_fp:
  install_requires = req_fp.readlines()

setup(
  name='joeynmt',
  version='0.0.2',
  description='Minimalist NMT for educational purposes',
  author='Jasmijn Bastings and Julia Kreutzer',
  url='https://github.com/joeynmt/joeynmt',
  license='Apache License',
  install_requires=install_requires,
  packages=find_packages(exclude=[]),
  python_requires='>=3.5',
  project_urls={
    'Documentation': 'http://joeynmt.readthedocs.io/en/latest/',
    'Source': 'https://github.com/joeynmt/joeynmt',
    'Tracker': 'https://github.com/joeynmt/joeynmt/issues',
  },
  entry_points={
    'console_scripts': [
    ],
  }
)



# coding: utf-8
"""
This module holds various MT evaluation metrics.
"""

import sacrebleu


def chrf(hypotheses, references):
    """
    Character F-score from sacrebleu

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return:
    """
    return sacrebleu.corpus_chrf(hypotheses=hypotheses, references=references)


def bleu(hypotheses, references):
    """
    Raw corpus BLEU from sacrebleu (without tokenization)

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return:
    """
    return sacrebleu.raw_corpus_bleu(sys_stream=hypotheses,
                                     ref_streams=[references]).score


def token_accuracy(hypotheses, references, level="word"):
    """
    Compute the accuracy of hypothesis tokens: correct tokens / all tokens
    Tokens are correct if they appear in the same position in the reference.

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :param level: segmentation level, either "word", "bpe", or "char"
    :return:
    """
    def split_by_space(string):
        """
        Helper method to split the input based on spaces.
        Follows the same structure as list(inp)
        :param string: string
        :return: list of strings
        """
        return string.split(" ")

    correct_tokens = 0
    all_tokens = 0
    split_func = split_by_space if level in ["word", "bpe"] else list
    assert len(hypotheses) == len(references)
    for hyp, ref in zip(hypotheses, references):
        all_tokens += len(hyp)
        for h_i, r_i in zip(split_func(hyp), split_func(ref)):
            # min(len(h), len(r)) tokens considered
            if h_i == r_i:
                correct_tokens += 1
    return (correct_tokens / all_tokens)*100 if all_tokens > 0 else 0.0


def sequence_accuracy(hypotheses, references):
    """
    Compute the accuracy of hypothesis tokens: correct tokens / all tokens
    Tokens are correct if they appear in the same position in the reference.

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return:
    """
    assert len(hypotheses) == len(references)
    correct_sequences = sum([1 for (hyp, ref) in zip(hypotheses, references)
                             if hyp == ref])
    return (correct_sequences / len(hypotheses))*100 if hypotheses else 0.0



import torch
import numpy as np

from joeynmt.search import greedy, recurrent_greedy, transformer_greedy
from joeynmt.search import beam_search
from joeynmt.decoders import RecurrentDecoder, TransformerDecoder
from joeynmt.encoders import RecurrentEncoder
from joeynmt.embeddings import Embeddings

from .test_helpers import TensorTestCase


# TODO for transformer and rnn, make sure both return the same result for
# beam_size<2 and greedy decoding

class TestSearch(TensorTestCase):
    def setUp(self):
        self.emb_size = 12
        self.num_layers = 3
        self.hidden_size = 12
        self.ff_size = 24
        self.num_heads = 4
        self.dropout = 0.
        self.encoder_hidden_size = 3
        self.vocab_size = 5
        seed = 42
        torch.manual_seed(seed)
        self.bos_index = 2
        self.pad_index = 1
        self.eos_index = 3

class TestSearchTransformer(TestSearch):

    def _build(self, batch_size):
        src_time_dim = 4
        vocab_size = 7

        emb = Embeddings(embedding_dim=self.emb_size, vocab_size=vocab_size,
                         padding_idx=self.pad_index)

        decoder = TransformerDecoder(
            num_layers=self.num_layers, num_heads=self.num_heads,
            hidden_size=self.hidden_size, ff_size=self.ff_size,
            dropout=self.dropout, emb_dropout=self.dropout,
            vocab_size=vocab_size)

        encoder_output = torch.rand(
            size=(batch_size, src_time_dim, self.hidden_size))

        for p in decoder.parameters():
            torch.nn.init.uniform_(p, -0.5, 0.5)

        src_mask = torch.ones(size=(batch_size, 1, src_time_dim)) == 1

        encoder_hidden = None  # unused
        return src_mask, emb, decoder, encoder_output, encoder_hidden

    def test_transformer_greedy(self):
        batch_size = 2
        max_output_length = 3
        src_mask, embed, decoder, encoder_output, \
        encoder_hidden = self._build(batch_size=batch_size)
        output, attention_scores = transformer_greedy(
            src_mask=src_mask, embed=embed, bos_index=self.bos_index,
            eos_index=self.eos_index,
            max_output_length=max_output_length, decoder=decoder,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)
        # Transformer greedy doesn't return attention scores
        self.assertIsNone(attention_scores)
        # batch x time
        self.assertEqual(output.shape, (batch_size, max_output_length))
        np.testing.assert_equal(output, [[5, 5, 5], [5, 5, 5]])

    def test_transformer_beam1(self):
        batch_size = 2
        beam_size = 1
        alpha = 1.
        max_output_length = 3
        src_mask, embed, decoder, encoder_output, \
        encoder_hidden = self._build(batch_size=batch_size)
        output, attention_scores = beam_search(
            size=beam_size, eos_index=self.eos_index, pad_index=self.pad_index,
            src_mask=src_mask, embed=embed, bos_index=self.bos_index,
            max_output_length=max_output_length, decoder=decoder, alpha=alpha,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)
        # Transformer beam doesn't return attention scores
        self.assertIsNone(attention_scores)
        # batch x time
        self.assertEqual(output.shape, (batch_size, max_output_length))
        np.testing.assert_equal(output, [[5, 5, 5], [5, 5, 5]])

        # now compare to greedy, they should be the same for beam=1
        greedy_output, _ = transformer_greedy(
            src_mask=src_mask, embed=embed, bos_index=self.bos_index,
            eos_index=self.eos_index,
            max_output_length=max_output_length, decoder=decoder,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)
        np.testing.assert_equal(output, greedy_output)

    def test_transformer_beam7(self):
        batch_size = 2
        beam_size = 7
        alpha = 1.
        max_output_length = 3
        src_mask, embed, decoder, encoder_output, \
        encoder_hidden = self._build(batch_size=batch_size)
        output, attention_scores = beam_search(
            size=beam_size, eos_index=self.eos_index, pad_index=self.pad_index,
            src_mask=src_mask, embed=embed, bos_index=self.bos_index, n_best=1,
            max_output_length=max_output_length, decoder=decoder, alpha=alpha,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)
        # Transformer beam doesn't return attention scores
        self.assertIsNone(attention_scores)
        # batch x time
        # now it produces EOS, so everything after gets cut off
        self.assertEqual(output.shape, (batch_size, 1))
        np.testing.assert_equal(output, [[3], [3]])


class TestSearchRecurrent(TestSearch):
    def _build(self, batch_size):
        src_time_dim = 4
        vocab_size = 7

        emb = Embeddings(embedding_dim=self.emb_size, vocab_size=vocab_size,
                         padding_idx=self.pad_index)

        encoder = RecurrentEncoder(emb_size=self.emb_size,
                                   num_layers=self.num_layers,
                                   hidden_size=self.encoder_hidden_size,
                                   bidirectional=True)

        decoder = RecurrentDecoder(hidden_size=self.hidden_size,
                                   encoder=encoder, attention="bahdanau",
                                   emb_size=self.emb_size,
                                   vocab_size=self.vocab_size,
                                   num_layers=self.num_layers,
                                   init_hidden="bridge",
                                   input_feeding=True)

        encoder_output = torch.rand(
            size=(batch_size, src_time_dim, encoder.output_size))

        for p in decoder.parameters():
            torch.nn.init.uniform_(p, -0.5, 0.5)

        src_mask = torch.ones(size=(batch_size, 1, src_time_dim)) == 1

        encoder_hidden = torch.rand(size=(batch_size, encoder.output_size))

        return src_mask, emb, decoder, encoder_output, encoder_hidden

    def test_recurrent_greedy(self):
        batch_size = 2
        max_output_length = 3
        src_mask, emb, decoder, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        output, attention_scores = recurrent_greedy(
            src_mask=src_mask, embed=emb, bos_index=self.bos_index,
            eos_index=self.eos_index, max_output_length=max_output_length,
            decoder=decoder,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)

        self.assertEqual(output.shape, (batch_size, max_output_length))
        np.testing.assert_equal(output, [[4, 0, 4], [4, 4, 4]])

        expected_attention_scores = np.array(
            [[[0.22914883, 0.24638498, 0.21247596, 0.3119903],
              [0.22970565, 0.24540883, 0.21261126, 0.31227428],
              [0.22903332, 0.2459198,  0.2110187,  0.3140282]],
             [[0.252522, 0.29074305, 0.257121, 0.19961396],
              [0.2519883,  0.2895494,  0.25718424, 0.201278],
              [0.2523954,  0.28959078, 0.25769445, 0.2003194]]])
        np.testing.assert_array_almost_equal(attention_scores,
                                             expected_attention_scores)
        self.assertEqual(attention_scores.shape, (batch_size, max_output_length,
                                                  4))

    def test_recurrent_beam1(self):
        # beam=1 and greedy should return the same result
        batch_size = 2
        max_output_length = 3
        src_mask, emb, decoder, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        greedy_output, _ = recurrent_greedy(
            src_mask=src_mask, embed=emb, bos_index=self.bos_index,
            eos_index=self.eos_index,
            max_output_length=max_output_length, decoder=decoder,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)

        beam_size = 1
        alpha = 1.0
        output, _ = beam_search(
            size=beam_size, eos_index=self.eos_index, pad_index=self.pad_index,
            src_mask=src_mask, embed=emb, bos_index=self.bos_index, n_best=1,
            max_output_length=max_output_length, decoder=decoder, alpha=alpha,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)
        np.testing.assert_array_equal(greedy_output, output)

    def test_recurrent_beam7(self):
        batch_size = 2
        max_output_length = 3
        src_mask, emb, decoder, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        beam_size = 7
        alpha = 1.0
        output, _ = beam_search(
            size=beam_size, eos_index=self.eos_index, pad_index=self.pad_index,
            src_mask=src_mask, embed=emb, bos_index=self.bos_index, n_best=1,
            max_output_length=max_output_length, decoder=decoder, alpha=alpha,
            encoder_output=encoder_output, encoder_hidden=encoder_hidden)

        self.assertEqual(output.shape, (2, 1))
        np.testing.assert_array_equal(output, [[3], [3]])
